# coding=utf-8
from __future__ import division

from ladybug._datacollectionbase import BaseCollection
from ladybug.datacollection import HourlyDiscontinuousCollection, \
    HourlyContinuousCollection, MonthlyCollection, DailyCollection, \
    MonthlyPerHourCollection
from ladybug.header import Header
from ladybug.analysisperiod import AnalysisPeriod
from ladybug.dt import DateTime
from ladybug.datatype.generic import GenericType
from ladybug.datatype.temperature import Temperature
from ladybug.datatype.fraction import RelativeHumidity, HumidityRatio

from ladybug.epw import EPW
from ladybug.psychrometrics import humid_ratio_from_db_rh

import pytest
import sys
if (sys.version_info >= (3, 0)):
    xrange = range


def test_init():
    """Test the init methods for base collections."""
    a_per = AnalysisPeriod(6, 21, 12, 6, 21, 13)
    dt1, dt2 = DateTime(6, 21, 12), DateTime(6, 21, 13)
    v1, v2 = 20, 25
    avg = (v1 + v2) / 2
    # Setup data collection
    dc1 = BaseCollection(Header(Temperature(), 'C', a_per), [v1, v2], [dt1, dt2])

    assert dc1.datetimes == (dt1, dt2)
    assert dc1.values == (v1, v2)
    assert dc1.average == avg
    str(dc1)  # Test the string representation of the collection
    str(dc1.header)  # Test the string representation of the header


def test_init_incorrect():
    """Test the init methods for base collections with incorrect inputs."""
    a_per = AnalysisPeriod(6, 21, 12, 6, 21, 13)
    dt1, dt2 = DateTime(6, 21, 12), DateTime(6, 21, 13)
    v1, v2 = 20, 25
    avg = (v1 + v2) / 2

    dc1 = BaseCollection(Header(Temperature(), 'C', a_per), [v1, v2], [dt1, dt2])
    dc1 = BaseCollection(Header(Temperature(), 'C', a_per), [v1, v2], (dt1, dt2))
    dc1 = BaseCollection(Header(Temperature(), 'C', a_per), [v1, v2], xrange(2))
    assert dc1.average == avg
    with pytest.raises(Exception):
        dc1 = BaseCollection(Header(Temperature(), 'C', a_per), [v1, v2], 'te')

    with pytest.raises(Exception):
        dc1 = BaseCollection(Header(Temperature(), 'C', a_per),
                             [v1, v2], {'1': 1, '2': 2})


def test_init_hourly():
    """Test the init methods for dicontinuous collections."""
    a_per = AnalysisPeriod(6, 21, 12, 6, 21, 13)
    dt1, dt2 = DateTime(6, 21, 12), DateTime(6, 21, 13)
    v1, v2 = 20, 25
    avg = (v1 + v2) / 2
    # Setup data collection
    dc1 = HourlyDiscontinuousCollection(Header(Temperature(), 'C', a_per),
                                        [v1, v2], [dt1, dt2])

    assert dc1.datetimes == (dt1, dt2)
    assert dc1.values == (v1, v2)
    assert dc1.average == avg
    assert not dc1.is_continuous
    str(dc1)  # Test the string representation of the collection
    str(dc1.header)  # Test the string representation of the header


def test_init_daily():
    """Test the init methods for daily collections."""
    a_per = AnalysisPeriod(6, 21, 0, 6, 22, 23)
    v1, v2 = 20, 25
    avg = (v1 + v2) / 2
    # Setup data collection
    dc1 = DailyCollection(Header(Temperature(), 'C', a_per),
                          [v1, v2], a_per.doys_int)

    assert dc1.datetimes == tuple(a_per.doys_int)
    assert dc1.values == (v1, v2)
    assert dc1.average == avg
    assert not dc1.is_continuous
    str(dc1)  # Test the string representation of the collection
    str(dc1.header)  # Test the string representation of the header


def test_init_monthly():
    """Test the init methods for monthly collections."""
    a_per = AnalysisPeriod(6, 1, 0, 7, 31, 23)
    v1, v2 = 20, 25
    avg = (v1 + v2) / 2
    # Setup data collection
    dc1 = MonthlyCollection(Header(Temperature(), 'C', a_per),
                            [v1, v2], a_per.months_int)

    assert dc1.datetimes == tuple(a_per.months_int)
    assert dc1.values == (v1, v2)
    assert dc1.average == avg
    assert not dc1.is_continuous
    str(dc1)  # Test the string representation of the collection
    str(dc1.header)  # Test the string representation of the header


def test_init_monthly_per_hour():
    """Test the init methods for monthly per hour collections."""
    a_per = AnalysisPeriod(6, 1, 0, 7, 31, 23)
    vals = [20] * 24 + [25] * 24
    avg = sum(vals) / 48
    # Setup data collection
    dc1 = MonthlyPerHourCollection(Header(Temperature(), 'C', a_per),
                                   vals, a_per.months_per_hour)

    assert dc1.datetimes == tuple(a_per.months_per_hour)
    assert dc1.values == tuple(vals)
    assert dc1.average == avg
    assert not dc1.is_continuous
    str(dc1)  # Test the string representation of the collection
    str(dc1.header)  # Test the string representation of the header


def test_init_continuous():
    """Test the init methods for continuous collections"""
    # Setup temperature data collection
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc1 = HourlyContinuousCollection(header, values)

    assert len(dc1.datetimes) == 8760
    assert list(dc1.values) == list(xrange(8760))
    assert dc1.average == 4379.5
    assert dc1.is_continuous
    str(dc1)  # Test the string representation of the collection
    str(dc1.header)  # Test the string representation of the header


def test_init_continuous_incorrect():
    """Test the init methods for continuous collections with incorrect values"""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(10))
    with pytest.raises(Exception):
        HourlyContinuousCollection(header, values)


def test_operators_hourly_discontinuous():
    """Test the operators for dicontinuous collections."""
    a_per = AnalysisPeriod(6, 21, 12, 6, 21, 13)
    dt1, dt2 = DateTime(6, 21, 12), DateTime(6, 21, 13)
    v1, v2 = 20, 25
    dc1 = HourlyDiscontinuousCollection(Header(Temperature(), 'C', a_per),
                                        [v1, v2], [dt1, dt2])
    dc2 = HourlyDiscontinuousCollection(Header(Temperature(), 'C', a_per),
                                        [v2, v1], [dt1, dt2])

    add = dc1 + dc2
    assert isinstance(add, HourlyDiscontinuousCollection)
    assert add.values == (v1 + v2, v2 + v1)

    sub = dc1 - dc2
    assert isinstance(sub, HourlyDiscontinuousCollection)
    assert sub.values == (v1 - v2, v2 - v1)

    mul = dc1 * dc2
    assert isinstance(mul, HourlyDiscontinuousCollection)
    assert mul.values == (v1 * v2, v2 * v1)

    div = dc1 / dc2
    assert isinstance(div, HourlyDiscontinuousCollection)
    assert div.values == (v1 / v2, v2 / v1)

    add = dc1 + 2
    assert isinstance(add, HourlyDiscontinuousCollection)
    assert add.values == (v1 + 2, v2 + 2)

    sub = dc1 - 2
    assert isinstance(sub, HourlyDiscontinuousCollection)
    assert sub.values == (v1 - 2, v2 - 2)

    mul = dc1 * 2
    assert isinstance(mul, HourlyDiscontinuousCollection)
    assert mul.values == (v1 * 2, v2 * 2)

    div = dc1 / 2
    assert isinstance(div, HourlyDiscontinuousCollection)
    assert div.values == (v1 / 2, v2 / 2)

    neg = -dc1
    assert isinstance(neg, HourlyDiscontinuousCollection)
    assert neg.values == (-v1, -v2)


def test_operators_hourly_continuous():
    """Test the operators for dicontinuous collections."""
    v1 = 20
    vals = [v1] * 24
    a_per = AnalysisPeriod(6, 21, 0, 6, 21, 23)
    # Setup data collection
    dc1 = HourlyContinuousCollection(Header(Temperature(), 'C', a_per), vals)
    dc2 = HourlyContinuousCollection(Header(Temperature(), 'C', a_per), vals)

    add = dc1 + dc2
    assert isinstance(add, HourlyContinuousCollection)
    assert add[0] == v1 + v1

    sub = dc1 - dc2
    assert isinstance(sub, HourlyContinuousCollection)
    assert sub[0] == v1 - v1

    mul = dc1 * dc2
    assert isinstance(mul, HourlyContinuousCollection)
    assert mul[0] == v1 * v1

    div = dc1 / dc2
    assert isinstance(div, HourlyContinuousCollection)
    assert div[0] == v1 / v1

    add = dc1 + 2
    assert isinstance(add, HourlyContinuousCollection)
    assert add[0] == v1 + 2

    sub = dc1 - 2
    assert isinstance(sub, HourlyContinuousCollection)
    assert sub[0] == v1 - 2

    mul = dc1 * 2
    assert isinstance(mul, HourlyContinuousCollection)
    assert mul[0] == v1 * 2

    div = dc1 / 2
    assert isinstance(div, HourlyContinuousCollection)
    assert div[0] == v1 / 2

    neg = -dc1
    assert isinstance(neg, HourlyContinuousCollection)
    assert neg[0] == -v1


def test_setting_values():
    """Test the methods for setting values on the data collection"""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyDiscontinuousCollection(header, values,
                                       header.analysis_period.datetimes)
    # test the contains and reversed methods
    assert 10 in dc
    assert list(reversed(dc)) == list(reversed(values))

    # test setting individual values
    assert dc[0] == 0
    dc[0] = 10
    assert dc[0] == 10

    # try setting the whole list of values
    val_rev = list(reversed(values))
    dc.values = val_rev
    assert dc[0] == 8759

    # make sure that people can't change the values without changing datetimes
    with pytest.raises(Exception):
        dc.values.append(10)


def test_setting_values_continuous():
    """Test the methods for setting values on the continuous data collection"""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyContinuousCollection(header, values)

    # try setting the whole list of values
    assert dc[0] == 0
    val_rev = list(reversed(values))
    dc.values = val_rev
    assert dc[0] == 8759

    # make sure that people can't change the number of values
    with pytest.raises(Exception):
        dc.values.append(10)


def test_validate_a_period_hourly():
    """Test the validate_analysis_period method for dicontinuous collections."""
    a_per = AnalysisPeriod(6, 21, 0, 6, 21, 23)
    dt1, dt2 = DateTime(6, 21, 12), DateTime(6, 21, 13)
    v1, v2 = 20, 25

    # Test that the validate method correctly sorts reversed datetimes.
    dc1 = HourlyDiscontinuousCollection(Header(Temperature(), 'C', a_per),
                                        [v1, v2], [dt2, dt1])
    dc1_new = dc1.validate_analysis_period()
    assert not dc1.validated_a_period
    assert dc1_new.validated_a_period
    assert dc1.datetimes == (dt2, dt1)
    assert dc1_new.datetimes == (dt1, dt2)

    # Test that the validate method correctly updates analysis_period range.
    a_per_2 = AnalysisPeriod(6, 20, 15, 6, 20, 23)
    dc1 = HourlyDiscontinuousCollection(Header(Temperature(), 'C', a_per_2),
                                        [v1, v2], [dt1, dt2])
    dc1_new = dc1.validate_analysis_period()
    assert not dc1.validated_a_period
    assert dc1_new.validated_a_period
    assert dc1.header.analysis_period == a_per_2
    assert dc1_new.header.analysis_period == AnalysisPeriod(
        6, 20, 12, 6, 21, 23)

    # Test that the validate method with reversed analysis_periods.
    a_per_3 = AnalysisPeriod(6, 20, 15, 2, 20, 23)
    dt5 = DateTime(1, 21, 12)
    dc1 = HourlyDiscontinuousCollection(Header(Temperature(), 'C', a_per_3),
                                        [v1, v2, v2], [dt1, dt2, dt5])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.header.analysis_period == AnalysisPeriod(
        6, 20, 12, 2, 20, 23)
    dc1 = HourlyDiscontinuousCollection(Header(Temperature(), 'C', a_per_3),
                                        [v1, v2], [dt1, dt2])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.header.analysis_period == AnalysisPeriod(
        6, 20, 12, 2, 20, 23)
    dc1 = HourlyDiscontinuousCollection(Header(Temperature(), 'C', a_per_3),
                                        [v1, v2], [dt5, DateTime(1, 21, 15)])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.header.analysis_period == AnalysisPeriod(
        6, 20, 12, 2, 20, 23)

    # Test that the validate method correctly updates timestep.
    dt3 = DateTime(6, 21, 12, 30)
    dc1 = HourlyDiscontinuousCollection(Header(Temperature(), 'C', a_per),
                                        [v1, v2, v2], [dt1, dt3, dt2])
    dc1_new = dc1.validate_analysis_period()
    assert not dc1.validated_a_period
    assert dc1_new.validated_a_period
    assert dc1.header.analysis_period.timestep == 1
    assert dc1_new.header.analysis_period.timestep == 2

    # Test that the validate method correctly identifies leap years.
    dt4 = DateTime(2, 29, 12, leap_year=True)
    dc1 = HourlyDiscontinuousCollection(Header(Temperature(), 'C', a_per),
                                        [v1, v2, v2], [dt1, dt4, dt2])
    dc1_new = dc1.validate_analysis_period()
    assert not dc1.validated_a_period
    assert dc1_new.validated_a_period
    assert not dc1.header.analysis_period.is_leap_year
    assert dc1_new.header.analysis_period.is_leap_year

    # Test that duplicated datetimes are caught
    dc1 = HourlyDiscontinuousCollection(Header(Temperature(), 'C', a_per),
                                        [v1, v2], [dt1, dt1])
    with pytest.raises(Exception):
        dc1_new = dc1.validate_analysis_period()


def test_validate_a_period_daily():
    """Test the validate_a_period methods for daily collections."""
    a_per = AnalysisPeriod(6, 21, 0, 6, 22, 23)
    v1, v2 = 20, 25
    dt1, dt2 = 172, 173

    # Test that the validate method correctly sorts reversed datetimes.
    dc1 = DailyCollection(Header(Temperature(), 'C', a_per),
                          [v1, v2], [dt2, dt1])
    dc1_new = dc1.validate_analysis_period()
    assert not dc1.validated_a_period
    assert dc1_new.validated_a_period
    assert dc1.datetimes == (dt2, dt1)
    assert dc1_new.datetimes == (dt1, dt2)

    # Test that the validate method correctly updates analysis_period range.
    a_per_2 = AnalysisPeriod(6, 20, 0, 6, 20, 23)
    dc1 = DailyCollection(Header(Temperature(), 'C', a_per_2),
                          [v1, v2], [dt1, dt2])
    dc1_new = dc1.validate_analysis_period()
    assert not dc1.validated_a_period
    assert dc1_new.validated_a_period
    assert dc1.header.analysis_period == a_per_2
    assert dc1_new.header.analysis_period == AnalysisPeriod(
        6, 20, 0, 6, 22, 23)

    # Test that the validate method with reversed analysis_periods.
    a_per_3 = AnalysisPeriod(6, 20, 0, 2, 20, 23)
    dt5 = 21
    dc1 = DailyCollection(Header(Temperature(), 'C', a_per_3),
                          [v1, v2, v2], [dt1, dt2, dt5])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.header.analysis_period == a_per_3
    dc1 = DailyCollection(Header(Temperature(), 'C', a_per_3),
                          [v1, v2], [dt1, dt2])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.header.analysis_period == a_per_3
    dc1 = DailyCollection(Header(Temperature(), 'C', a_per_3),
                          [v1, v2], [dt5, 22])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.header.analysis_period == a_per_3
    dc1 = DailyCollection(Header(Temperature(), 'C', a_per_3),
                          [v1, v2], [dt5, 60])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.header.analysis_period == AnalysisPeriod()

    # Test that the validate method correctly identifies leap years.
    dc1 = DailyCollection(Header(Temperature(), 'C', a_per),
                          [v1, v2, v2], [dt1, dt2, 366])
    dc1_new = dc1.validate_analysis_period()
    assert not dc1.validated_a_period
    assert dc1_new.validated_a_period
    assert not dc1.header.analysis_period.is_leap_year
    assert dc1_new.header.analysis_period.is_leap_year

    # Test that duplicated datetimes are caught
    dc1 = DailyCollection(Header(Temperature(), 'C', a_per),
                          [v1, v2], [dt1, dt1])
    with pytest.raises(Exception):
        dc1_new = dc1.validate_analysis_period()


def test_validate_a_period_monthly():
    """Test the validate_a_period methods for monthly collections."""
    a_per = AnalysisPeriod(6, 1, 0, 7, 1, 23)
    v1, v2 = 20, 25
    dt1, dt2 = 6, 7

    # Test that the validate method correctly sorts reversed datetimes.
    dc1 = MonthlyCollection(Header(Temperature(), 'C', a_per),
                            [v1, v2], [dt2, dt1])
    dc1_new = dc1.validate_analysis_period()
    assert not dc1.validated_a_period
    assert dc1_new.validated_a_period
    assert dc1.datetimes == (dt2, dt1)
    assert dc1_new.datetimes == (dt1, dt2)

    # Test that the validate method correctly updates analysis_period range.
    a_per_2 = AnalysisPeriod(6, 1, 0, 6, 1, 23)
    dc1 = MonthlyCollection(Header(Temperature(), 'C', a_per_2),
                            [v1, v2], [dt1, dt2])
    dc1_new = dc1.validate_analysis_period()
    assert not dc1.validated_a_period
    assert dc1_new.validated_a_period
    assert dc1.header.analysis_period == a_per_2
    assert dc1_new.header.analysis_period == AnalysisPeriod(
        6, 1, 0, 7, 31, 23)

    # Test that the validate method with reversed analysis_periods.
    a_per_3 = AnalysisPeriod(6, 1, 0, 2, 28, 23)
    dt5 = 1
    dc1 = MonthlyCollection(Header(Temperature(), 'C', a_per_3),
                            [v1, v2, v2], [dt5, dt1, dt2])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.header.analysis_period == a_per_3
    assert dc1_new.datetimes == (dt1, dt2, dt5)
    dc1 = MonthlyCollection(Header(Temperature(), 'C', a_per_3),
                            [v1, v2], [dt1, dt2])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.header.analysis_period == a_per_3
    assert dc1_new.datetimes == (dt1, dt2)
    dc1 = MonthlyCollection(Header(Temperature(), 'C', a_per_3),
                            [v1, v2], [dt5, 2])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.datetimes == (dt5, 2)
    assert dc1_new.header.analysis_period == a_per_3
    dc1 = MonthlyCollection(Header(Temperature(), 'C', a_per_3),
                            [v1, v2], [dt5, 4])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.header.analysis_period == AnalysisPeriod()
    assert dc1_new.datetimes == (dt5, 4)

    # Test that duplicated datetimes are caught
    dc1 = MonthlyCollection(Header(Temperature(), 'C', a_per),
                            [v1, v2], [dt1, dt1])
    with pytest.raises(Exception):
        dc1_new = dc1.validate_analysis_period()


def test_validate_a_period_monthly_per_hour():
    """Test validate_a_period for monthly_per_hour collections."""
    a_per = AnalysisPeriod(6, 1, 0, 7, 1, 23)
    v1, v2 = 20, 25
    dt1, dt2 = (6, 12), (7, 13)

    # Test that the validate method correctly sorts reversed datetimes.
    dc1 = MonthlyPerHourCollection(Header(Temperature(), 'C', a_per),
                                   [v1, v2], [dt2, dt1])
    dc1_new = dc1.validate_analysis_period()
    assert not dc1.validated_a_period
    assert dc1_new.validated_a_period
    assert dc1.datetimes == (dt2, dt1)
    assert dc1_new.datetimes == (dt1, dt2)

    # Test that the validate method correctly updates analysis_period range.
    a_per_2 = AnalysisPeriod(6, 1, 15, 6, 1, 23)
    dc1 = MonthlyPerHourCollection(Header(Temperature(), 'C', a_per_2),
                                   [v1, v2], [dt1, dt2])
    dc1_new = dc1.validate_analysis_period()
    assert not dc1.validated_a_period
    assert dc1_new.validated_a_period
    assert dc1.header.analysis_period == a_per_2
    assert dc1_new.header.analysis_period == AnalysisPeriod(
        6, 1, 12, 7, 31, 23)

    # Test that the validate method with reversed analysis_periods.
    a_per_3 = AnalysisPeriod(6, 1, 0, 2, 28, 23)
    dt5 = (1, 15)
    dc1 = MonthlyPerHourCollection(Header(Temperature(), 'C', a_per_3),
                                   [v1, v2, v2], [dt5, dt1, dt2])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.header.analysis_period == a_per_3
    assert dc1_new.datetimes == (dt1, dt2, dt5)
    dc1 = MonthlyPerHourCollection(Header(Temperature(), 'C', a_per_3),
                                   [v1, v2], [dt1, dt2])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.header.analysis_period == a_per_3
    assert dc1_new.datetimes == (dt1, dt2)
    dc1 = MonthlyPerHourCollection(Header(Temperature(), 'C', a_per_3),
                                   [v1, v2], [dt5, (2, 12)])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.datetimes == (dt5, (2, 12))
    assert dc1_new.header.analysis_period == a_per_3
    dc1 = MonthlyPerHourCollection(Header(Temperature(), 'C', a_per_3),
                                   [v1, v2], [dt5, (4, 12)])
    dc1_new = dc1.validate_analysis_period()
    assert dc1_new.header.analysis_period == AnalysisPeriod()
    assert dc1_new.datetimes == (dt5, (4, 12))

    # Test that duplicated datetimes are caught
    dc1 = MonthlyPerHourCollection(Header(Temperature(), 'C', a_per),
                                   [v1, v2], [dt1, dt1])
    with pytest.raises(Exception):
        dc1_new = dc1.validate_analysis_period()


def test_bounds():
    """Test the bounds method."""
    header1 = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc1 = HourlyContinuousCollection(header1, values)
    min, max = dc1.bounds
    assert min == 0
    assert max == 8759


def test_min_max():
    """Test the min and max methods."""
    header1 = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(10, 8770))
    dc = HourlyContinuousCollection(header1, values)
    assert dc.min == 10
    assert dc.max == 8769


def test_average_median():
    """Test the average and median methods."""
    header1 = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyContinuousCollection(header1, values)
    assert dc.average == dc.median == 4379.5
    dc[0] = -10000
    assert dc.average == pytest.approx(4378.35844, rel=1e-3)
    assert dc.median == 4379.5


def test_total():
    """Test the total method."""
    header1 = Header(Temperature(), 'C', AnalysisPeriod())
    values = [1] * 8760
    dc = HourlyContinuousCollection(header1, values)
    assert dc.total == 8760


def test_to_unit():
    """Test the conversion of DataCollection units."""
    header1 = Header(Temperature(), 'C', AnalysisPeriod())
    header2 = Header(Temperature(), 'F', AnalysisPeriod())
    values = [20] * 8760
    dc1 = HourlyContinuousCollection(header1, values)
    dc2 = HourlyContinuousCollection(header2, values)
    dc3 = dc1.to_unit('K')
    dc4 = dc2.to_unit('K')
    assert dc1.values[0] == 20
    assert dc3.values[0] == 293.15
    assert dc2.values[0] == 20
    assert dc4.values[0] == pytest.approx(266.483, rel=1e-1)


def test_to_ip_si():
    """Test the conversion of DataCollection to IP and SI units."""
    header1 = Header(Temperature(), 'C', AnalysisPeriod())
    values = [20] * 8760
    dc1 = HourlyContinuousCollection(header1, values)
    dc2 = dc1.to_ip()
    dc3 = dc2.to_si()
    assert dc1.values[0] == 20
    assert dc2.values[0] == 68
    assert dc3.values[0] == dc1.values[0]


def test_get_highest_values():
    header = Header(Temperature(), 'C', AnalysisPeriod())
    test_data = list(xrange(8760))
    test_count = len(test_data)/2
    dc3 = HourlyContinuousCollection(header, test_data)

    test_highest_values, test_highest_values_index = \
        dc3.get_highest_values(test_count)

    assert test_highest_values == list(reversed(test_data[4380:8760]))
    assert test_highest_values_index == list(xrange(8759, 4379, -1))


def test_get_lowest_values():
    header = Header(Temperature(), 'C', AnalysisPeriod())
    test_data = [50] * 8760
    test_count = len(test_data)/2
    dc3 = HourlyContinuousCollection(header, test_data)

    test_lowest_values, test_lowest_values_index = \
        dc3.get_lowest_values(test_count)

    assert test_lowest_values == list(test_data[0:4380])
    assert test_lowest_values_index == list(xrange(0, 4380))


def test_get_percentile():
    """Test the get_percentile method."""
    header1 = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyContinuousCollection(header1, values)

    assert dc.get_percentile(0) == 0
    assert dc.get_percentile(25) == 2189.75
    assert dc.get_percentile(50) == 4379.5
    assert dc.get_percentile(75) == 6569.25
    assert dc.get_percentile(100) == 8759

    with pytest.raises(Exception):
        dc.get_percentile(-10)
    with pytest.raises(Exception):
        dc.get_percentile(110)


def test_filter_by_conditional_statement():
    """Test filter by conditional statement."""
    a_per = AnalysisPeriod(end_month=1, end_day=2)
    header1 = Header(Temperature(), 'C', a_per)
    values = list(xrange(48))
    dc1 = HourlyDiscontinuousCollection(header1, values, a_per.datetimes)

    dc2 = dc1.filter_by_conditional_statement('a > 23')
    assert len(dc2) == 24
    for d in dc2.values:
        assert d > 23


def test_filter_by_conditional_statement_continuous():
    """Test filter by conditional statement."""
    header1 = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=2))
    values = list(xrange(48))
    dc1 = HourlyContinuousCollection(header1, values)

    dc2 = dc1.filter_by_conditional_statement('a > 23')
    assert len(dc2) == 24
    for d in dc2.values:
        assert d > 23
    assert not isinstance(dc2, HourlyContinuousCollection)


def test_filter_by_pattern():
    """Test filter by pattern."""
    a_per = AnalysisPeriod(end_month=1, end_day=2)
    header1 = Header(Temperature(), 'C', a_per)
    values = list(xrange(48))
    dc1 = HourlyDiscontinuousCollection(header1, values, a_per.datetimes)

    dc2 = dc1.filter_by_pattern([True, False] * 24)
    assert len(dc2) == 24


def test_filter_by_pattern_continuous():
    """Test filter by pattern."""
    header1 = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=2))
    values = list(xrange(48))
    dc1 = HourlyContinuousCollection(header1, values)

    dc2 = dc1.filter_by_pattern([True, False] * 24)
    assert len(dc2) == 24
    assert not isinstance(dc2, HourlyContinuousCollection)


def test_filter_by_analysis_period_sub_hourly():
    """Test filtering by analysis period on sub-hourly continuous collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(range(8760))
    dc1 = HourlyContinuousCollection(header, values)
    dc2 = dc1.interpolate_to_timestep(4)

    dc3 = dc2.filter_by_analysis_period(AnalysisPeriod(st_month=7, end_month=7, timestep=4))
    assert len(dc3) == 2976
    dc4 = dc2.filter_by_analysis_period(AnalysisPeriod(st_hour=9, end_hour=17, timestep=4))
    assert len(dc4) == 12045


def test_filter_by_analysis_period_hourly():
    """Test filtering by analysis period on hourly discontinuous collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyDiscontinuousCollection(header, values,
                                       header.analysis_period.datetimes)
    dc = dc.validate_analysis_period()
    a_per = AnalysisPeriod(st_month=3, end_month=3)
    filt_dc = dc.filter_by_analysis_period(a_per)
    assert len(filt_dc) == 31 * 24
    assert filt_dc.header.analysis_period == a_per
    assert filt_dc.datetimes[0] == DateTime(3, 1, 0)
    assert filt_dc.datetimes[-1] == DateTime(3, 31, 23)


def test_filter_by_analysis_period_daily():
    """Test filtering by analysis period on daily collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(365))
    dc = DailyCollection(header, values, AnalysisPeriod().doys_int)
    dc = dc.validate_analysis_period()
    a_per = AnalysisPeriod(st_month=3, end_month=4, end_day=30)
    filt_dc = dc.filter_by_analysis_period(a_per)
    assert len(filt_dc) == 31 + 30
    assert filt_dc.header.analysis_period == a_per
    assert filt_dc.datetimes[0] == 31 + 28 + 1
    assert filt_dc.datetimes[-1] == 31 + 28 + 31 + 30


def test_filter_by_analysis_period_monthly():
    """Test filtering by analysis period on monthly collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(12))
    dc = MonthlyCollection(header, values, AnalysisPeriod().months_int)
    a_per = AnalysisPeriod(st_month=3, end_month=6)
    filt_dc = dc.filter_by_analysis_period(a_per)
    assert len(filt_dc) == 4
    assert filt_dc.header.analysis_period == a_per
    assert filt_dc.datetimes[0] == 3
    assert filt_dc.datetimes[-1] == 6


def test_filter_by_analysis_period_monthly_per_hour():
    """Test filtering by analysis period on monthly per hour collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(12 * 24))
    dc = MonthlyPerHourCollection(header, values, AnalysisPeriod().months_per_hour)
    a_per = AnalysisPeriod(st_month=3, end_month=4, end_day=30)
    filt_dc = dc.filter_by_analysis_period(a_per)
    assert len(filt_dc) == 2 * 24
    assert filt_dc.header.analysis_period == a_per
    assert filt_dc.datetimes[0] == (3, 0)
    assert filt_dc.datetimes[-1] == (4, 23)


def test_filter_by_analysis_period_continuous():
    """Test filtering by analysis period on hourly continuous collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyContinuousCollection(header, values)
    a_per = AnalysisPeriod(st_month=3, end_month=3)
    filt_dc = dc.filter_by_analysis_period(a_per)
    assert len(filt_dc) == 31 * 24
    assert filt_dc.header.analysis_period == a_per
    assert filt_dc.datetimes[0] == DateTime(3, 1, 0)
    assert filt_dc.datetimes[-1] == DateTime(3, 31, 23)
    assert isinstance(filt_dc, HourlyContinuousCollection)


def test_filter_by_analysis_period_continuous_reversed():
    """Test filtering by reversed analysis period on hourly continuous collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyContinuousCollection(header, values)
    a_per = AnalysisPeriod(st_month=12, end_month=3)
    filt_dc = dc.filter_by_analysis_period(a_per)
    assert len(filt_dc) == (31 + 31 + 28 + 31) * 24
    assert filt_dc.header.analysis_period == a_per
    assert filt_dc.datetimes[0] == DateTime(12, 1, 0)
    assert filt_dc.datetimes[-1] == DateTime(3, 31, 23)
    assert isinstance(filt_dc, HourlyContinuousCollection)


def test_filter_by_analysis_period_continuous_hour_subset():
    """Test filtering hour subset analysis period on hourly continuous collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyContinuousCollection(header, values)
    a_per = AnalysisPeriod(3, 2, 9, 3, 8, 17)
    filt_dc = dc.filter_by_analysis_period(a_per)
    assert len(filt_dc) == 7 * 9
    assert filt_dc.header.analysis_period == a_per
    assert filt_dc.datetimes[0] == DateTime(3, 2, 9)
    assert filt_dc.datetimes[-1] == DateTime(3, 8, 17)
    assert not isinstance(filt_dc, HourlyContinuousCollection)


def test_filter_by_analysis_period_continuous_large():
    """Test filtering large analysis period on hourly continuous collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod(st_month=3, end_month=3))
    values = list(xrange(24 * 31))
    dc = HourlyContinuousCollection(header, values)
    a_per = AnalysisPeriod(st_hour=9, end_hour=17)
    filt_dc = dc.filter_by_analysis_period(a_per)
    assert len(filt_dc) == 31 * 9
    assert filt_dc.header.analysis_period == AnalysisPeriod(3, 1, 9, 3, 31, 17)
    assert filt_dc.datetimes[0] == DateTime(3, 1, 9)
    assert filt_dc.datetimes[-1] == DateTime(3, 31, 17)
    assert not isinstance(filt_dc, HourlyContinuousCollection)


def test_filter_by_hoys():
    """Test filter_by_hoys method."""
    a_per = AnalysisPeriod(st_month=3, end_month=3)
    header = Header(Temperature(), 'C', a_per)
    values = list(xrange(24 * 31))
    dc = HourlyDiscontinuousCollection(header, values, a_per.datetimes)
    hoys = AnalysisPeriod(st_hour=9, end_hour=17).hoys
    filt_dc = dc.filter_by_hoys(hoys)
    assert len(filt_dc) == 31 * 9
    assert filt_dc.datetimes[0] == DateTime(3, 1, 9)
    assert filt_dc.datetimes[-1] == DateTime(3, 31, 17)


def test_filter_by_hoys_continuous():
    """Test filter_by_hoys method."""
    header = Header(Temperature(), 'C', AnalysisPeriod(st_month=3, end_month=3))
    values = list(xrange(24 * 31))
    dc = HourlyContinuousCollection(header, values)
    hoys = AnalysisPeriod(st_hour=9, end_hour=17).hoys
    filt_dc = dc.filter_by_hoys(hoys)
    assert len(filt_dc) == 31 * 9
    assert filt_dc.datetimes[0] == DateTime(3, 1, 9)
    assert filt_dc.datetimes[-1] == DateTime(3, 31, 17)
    assert not isinstance(filt_dc, HourlyContinuousCollection)


def test_average_daily():
    """Test the average daily method."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyContinuousCollection(header, values)
    new_dc = dc.average_daily()
    assert isinstance(new_dc, DailyCollection)
    assert len(new_dc) == 365
    assert new_dc.datetimes[0] == 1
    assert new_dc.datetimes[-1] == 365
    assert new_dc.is_continuous
    for i, val in dc.group_by_day().items():
        assert new_dc[i - 1] == sum(val) / len(val)


def test_total_daily():
    """Test the total daily method."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyContinuousCollection(header, values)
    new_dc = dc.total_daily()
    assert isinstance(new_dc, DailyCollection)
    assert len(new_dc) == 365
    assert new_dc.datetimes[0] == 1
    assert new_dc.datetimes[-1] == 365
    assert new_dc.is_continuous
    for i, val in dc.group_by_day().items():
        assert new_dc[i - 1] == sum(val)


def test_percentile_daily():
    """Test the percentile daily method."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(24)) * 365
    dc = HourlyContinuousCollection(header, values)
    new_dc = dc.percentile_daily(25)
    assert isinstance(new_dc, DailyCollection)
    assert len(new_dc) == 365
    assert new_dc.datetimes[0] == 1
    assert new_dc.datetimes[-1] == 365
    assert new_dc.is_continuous
    for i, val in dc.group_by_day().items():
        assert new_dc[i - 1] == 5.75


def test_average_monthly():
    """Test the average monthly method."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyContinuousCollection(header, values)
    new_dc = dc.average_monthly()
    assert isinstance(new_dc, MonthlyCollection)
    assert len(new_dc) == 12
    assert new_dc.datetimes[0] == 1
    assert new_dc.datetimes[-1] == 12
    assert new_dc.is_continuous
    for i, val in dc.group_by_month().items():
        assert new_dc[i - 1] == sum(val) / len(val)


def test_total_monthly():
    """Test the total monthly method."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyContinuousCollection(header, values)
    new_dc = dc.total_monthly()
    assert isinstance(new_dc, MonthlyCollection)
    assert len(new_dc) == 12
    assert new_dc.datetimes[0] == 1
    assert new_dc.datetimes[-1] == 12
    assert new_dc.is_continuous
    for i, val in dc.group_by_month().items():
        assert new_dc[i - 1] == sum(val)


def test_percentile_monthly():
    """Test the percentile monthly method."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = [50] * 8760
    dc = HourlyContinuousCollection(header, values)
    new_dc = dc.percentile_monthly(25)
    assert isinstance(new_dc, MonthlyCollection)
    assert len(new_dc) == 12
    assert new_dc.datetimes[0] == 1
    assert new_dc.datetimes[-1] == 12
    assert new_dc.is_continuous
    for i, val in dc.group_by_month().items():
        assert new_dc[i - 1] == 50


def test_average_monthly_on_daily_collection():
    """Test the average monthly method."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(365))
    dc = DailyCollection(header, values, AnalysisPeriod().doys_int)
    new_dc = dc.average_monthly()
    assert isinstance(new_dc, MonthlyCollection)
    assert len(new_dc) == 12
    assert new_dc.datetimes[0] == 1
    assert new_dc.datetimes[-1] == 12
    for i, val in dc.group_by_month().items():
        assert new_dc[i - 1] == sum(val) / len(val)


def test_total_monthly_on_daily_collection():
    """Test the total monthly method."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(365))
    dc = DailyCollection(header, values, AnalysisPeriod().doys_int)
    new_dc = dc.total_monthly()
    assert isinstance(new_dc, MonthlyCollection)
    assert len(new_dc) == 12
    assert new_dc.datetimes[0] == 1
    assert new_dc.datetimes[-1] == 12
    for i, val in dc.group_by_month().items():
        assert new_dc[i - 1] == sum(val)


def test_percentile_monthly_on_daily_collection():
    """Test the percentile monthly method."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(365))
    dc = DailyCollection(header, values, AnalysisPeriod().doys_int)
    new_dc = dc.percentile_monthly(25)
    assert isinstance(new_dc, MonthlyCollection)
    assert len(new_dc) == 12
    assert new_dc.datetimes[0] == 1
    assert new_dc.datetimes[-1] == 12


def test_average_monthly_per_hour():
    """Test the average monthly per hour method."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyContinuousCollection(header, values)
    new_dc = dc.average_monthly_per_hour()
    assert isinstance(new_dc, MonthlyPerHourCollection)
    assert len(new_dc) == 12 * 24
    assert new_dc.datetimes[0] == (1, 0)
    assert new_dc.datetimes[-1] == (12, 23)
    assert new_dc.is_continuous
    for i, val in enumerate(dc.group_by_month_per_hour().values()):
        assert new_dc[i] == sum(val) / len(val)


def test_total_monthly_per_hour():
    """Test the total monthly per hour method."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyContinuousCollection(header, values)
    new_dc = dc.total_monthly_per_hour()
    assert isinstance(new_dc, MonthlyPerHourCollection)
    assert len(new_dc) == 12 * 24
    assert new_dc.datetimes[0] == (1, 0)
    assert new_dc.datetimes[-1] == (12, 23)
    assert new_dc.is_continuous
    for i, val in enumerate(dc.group_by_month_per_hour().values()):
        assert new_dc[i] == sum(val)


def test_percentile_monthly_per_hour():
    """Test the percentile monthly per hour method."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(24)) * 365
    dc = HourlyContinuousCollection(header, values)
    new_dc = dc.percentile_monthly_per_hour(25)
    assert isinstance(new_dc, MonthlyPerHourCollection)
    assert len(new_dc) == 12 * 24
    assert new_dc.datetimes[0] == (1, 0)
    assert new_dc.datetimes[-1] == (12, 23)
    assert new_dc.is_continuous
    pct_vals = list(xrange(24)) * 12
    for i, val in enumerate(pct_vals):
        assert new_dc[i] == val


def test_group_by_day_discontinuous():
    """Test the group by day method for dicontinuous collections."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyDiscontinuousCollection(header, values, AnalysisPeriod().datetimes)
    day_dict = dc.group_by_day()
    assert list(day_dict.keys()) == list(xrange(1, 366))
    for val in day_dict.values():
        assert len(val) == 24


def test_group_by_month_discontinuous():
    """Test the group by month method for dicontinuous collections."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc = HourlyDiscontinuousCollection(header, values, AnalysisPeriod().datetimes)
    day_dict = dc.group_by_month()
    assert list(day_dict.keys()) == list(xrange(1, 13))
    days_per_month = AnalysisPeriod().NUMOFDAYSEACHMONTH
    for i, val in enumerate(day_dict.values()):
        assert len(val) == 24 * days_per_month[i]


def test_interpolate_holes():
    """Test the interoplate holes method on the discontinuous collection."""
    a_per = AnalysisPeriod(6, 21, 0, 6, 21, 23)
    dt1, dt2 = DateTime(6, 21, 12), DateTime(6, 21, 14)
    v1, v2 = 20, 25
    dc1 = HourlyDiscontinuousCollection(Header(Temperature(), 'C', a_per),
                                        [v1, v2], [dt1, dt2])
    with pytest.raises(Exception):
        interp_coll1 = dc1.interpolate_holes()
    dc2 = dc1.validate_analysis_period()
    interp_coll1 = dc2.interpolate_holes()
    assert isinstance(interp_coll1, HourlyContinuousCollection)
    assert len(interp_coll1.values) == 24
    assert interp_coll1[0] == 20
    assert interp_coll1[12] == 20
    assert interp_coll1[13] == 22.5
    assert interp_coll1[14] == 25
    assert interp_coll1[23] == 25

    values = list(xrange(24))
    test_header = Header(GenericType('Test Type', 'test'), 'test',
                         AnalysisPeriod(end_month=1, end_day=1))
    dc3 = HourlyContinuousCollection(test_header, values)
    interp_coll2 = dc3.interpolate_holes()
    assert isinstance(interp_coll2, HourlyContinuousCollection)
    assert len(interp_coll2.values) == 24


def test_cull_to_timestep():
    """Test the test_cull_to_timestep method on the discontinuous collection."""
    a_per = AnalysisPeriod(6, 21, 0, 6, 21, 23)
    dt1, dt2, dt3 = DateTime(6, 21, 12), DateTime(6, 21, 13), DateTime(6, 21, 12, 30)
    v1, v2 = 20, 25
    dc1 = HourlyDiscontinuousCollection(Header(Temperature(), 'C', a_per),
                                        [v1, v2, v2], [dt1, dt3, dt2])
    dc1_new = dc1.validate_analysis_period()
    dc2_new = dc1.cull_to_timestep()
    assert dc1_new.header.analysis_period.timestep == 2
    assert dc2_new.header.analysis_period.timestep == 1
    assert len(dc1_new.values) == 3
    assert len(dc2_new.values) == 2

    dc1_new.convert_to_culled_timestep()
    assert dc1_new.header.analysis_period.timestep == 1
    assert len(dc1_new.values) == 2


def test_interpolate_to_timestep():
    """Test the interoplation method on the continuous collection."""
    values = list(xrange(24))
    test_header = Header(GenericType('Test Type', 'test'), 'test',
                         AnalysisPeriod(end_month=1, end_day=1))
    dc2 = HourlyContinuousCollection(test_header, values)

    # check the interpolate data functions
    interp_coll1 = dc2.interpolate_to_timestep(2)
    interp_coll2 = dc2.interpolate_to_timestep(2, True)
    assert interp_coll1[1] == 0.5
    assert interp_coll2[1] == 0.25
    assert 'Minute' in interp_coll1.timestep_text


def test_is_collection_aligned():
    """Test the test_is_collection_aligned method for discontinuous collections."""
    header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    header3 = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=2))
    header4 = Header(Temperature(), 'C', AnalysisPeriod(
        st_day=2, end_month=1, end_day=2))
    header5 = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=24))
    values1 = list(xrange(24))
    values2 = [12] * 24
    values3 = [12] * 48
    dc1 = HourlyDiscontinuousCollection(header, values1,
                                        header.analysis_period.datetimes)
    dc2 = HourlyDiscontinuousCollection(header, values2,
                                        header.analysis_period.datetimes)
    dc3 = HourlyDiscontinuousCollection(header3, values3,
                                        header3.analysis_period.datetimes)
    dc4 = HourlyDiscontinuousCollection(header4, values1,
                                        header4.analysis_period.datetimes)
    dc5 = DailyCollection(header5, values1, header5.analysis_period.doys_int)

    assert dc1.is_collection_aligned(dc2)
    assert dc2.is_collection_aligned(dc1)
    assert not dc1.is_collection_aligned(dc3)
    assert not dc3.is_collection_aligned(dc1)
    assert not dc1.is_collection_aligned(dc4)
    assert not dc4.is_collection_aligned(dc1)
    assert not dc1.is_collection_aligned(dc5)
    assert not dc5.is_collection_aligned(dc1)

    assert HourlyDiscontinuousCollection.are_collections_aligned([dc1, dc2])
    assert not HourlyDiscontinuousCollection.are_collections_aligned(
        [dc1, dc2, dc3], False)


def test_is_collection_aligned_continuous():
    """Test the test_is_collection_aligned method for continuous collections."""
    header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    header3 = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=2))
    values1 = list(xrange(24))
    values2 = [12] * 24
    values3 = [12] * 48
    dc1 = HourlyContinuousCollection(header, values1)
    dc2 = HourlyContinuousCollection(header, values2)
    dc3 = HourlyContinuousCollection(header3, values3)

    assert dc1.is_collection_aligned(dc2)
    assert dc2.is_collection_aligned(dc1)
    assert not dc1.is_collection_aligned(dc3)
    assert not dc3.is_collection_aligned(dc1)

    assert HourlyContinuousCollection.are_collections_aligned([dc1, dc2])
    assert not HourlyContinuousCollection.are_collections_aligned(
        [dc1, dc2, dc3], False)


def test_get_aligned_collection():
    """Test the method for getting an aligned discontinuous collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    values = list(xrange(24))
    dc1 = HourlyDiscontinuousCollection(header, values,
                                        header.analysis_period.datetimes)
    dc2 = dc1.get_aligned_collection(50)
    assert dc1.header.data_type.name == dc2.header.data_type.name
    assert dc1.header.unit == dc2.header.unit
    assert dc1.header.analysis_period == dc2.header.analysis_period
    assert dc1.header.metadata == dc2.header.metadata
    assert dc1.datetimes == dc2.datetimes
    assert len(dc1.values) == len(dc2.values)
    assert dc2.values == tuple([50] * 24)

    dc3 = dc1.get_aligned_collection(50, RelativeHumidity(), '%')
    assert dc3.header.data_type.name == 'Relative Humidity'
    assert dc3.header.unit == '%'


def test_get_aligned_collection_continuous():
    """Test the method for getting an aligned continuous collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    values = list(xrange(24))
    dc1 = HourlyContinuousCollection(header, values)
    dc2 = dc1.get_aligned_collection(50)
    assert dc1.header.data_type.name == dc2.header.data_type.name
    assert dc1.header.unit == dc2.header.unit
    assert dc1.header.analysis_period == dc2.header.analysis_period
    assert dc1.header.metadata == dc2.header.metadata
    assert len(dc1.values) == len(dc2.values)
    assert dc2.values == tuple([50] * 24)

    dc3 = dc1.get_aligned_collection(50, RelativeHumidity(), '%')
    assert dc3.header.data_type.name == 'Relative Humidity'
    assert dc3.header.unit == '%'


def test_compute_function_aligned():
    """Test the method for computing funtions with aligned collections."""
    epw_file_path = './tests/fixtures/epw/chicago.epw'
    chicago_epw = EPW(epw_file_path)
    pressure_at_chicago = 95000
    hr_inputs = [chicago_epw.dry_bulb_temperature,
                 chicago_epw.relative_humidity,
                 pressure_at_chicago]
    humid_ratio = HourlyContinuousCollection.compute_function_aligned(
        humid_ratio_from_db_rh, hr_inputs, HumidityRatio(), 'fraction')
    assert isinstance(humid_ratio, HourlyContinuousCollection)
    assert len(humid_ratio.values) == 8760
    for i, val in enumerate(humid_ratio.values):
        assert val == humid_ratio_from_db_rh(chicago_epw.dry_bulb_temperature[i],
                                             chicago_epw.relative_humidity[i],
                                             pressure_at_chicago)

    hr_inputs = [20, 70, pressure_at_chicago]
    humid_ratio = HourlyContinuousCollection.compute_function_aligned(
        humid_ratio_from_db_rh, hr_inputs, HumidityRatio(), 'fraction')
    assert isinstance(humid_ratio, float)
    assert humid_ratio == humid_ratio_from_db_rh(20, 70, pressure_at_chicago)


def test_duplicate():
    """Test the duplicate method on the discontinuous collections."""
    header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    values = list(xrange(24))
    dc1 = HourlyDiscontinuousCollection(header, values,
                                        header.analysis_period.datetimes)
    dc2 = dc1.duplicate()
    assert dc1.header.data_type.name == dc2.header.data_type.name
    assert dc1.header.unit == dc2.header.unit
    assert dc1.header.analysis_period == dc2.header.analysis_period
    assert dc1.header.metadata == dc2.header.metadata
    assert dc1.values == dc2.values
    assert dc1.datetimes == dc2.datetimes


def test_duplicate_continuous():
    """Test the duplicate method on the continuous collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    values = list(xrange(24))
    dc1 = HourlyContinuousCollection(header, values)
    dc2 = dc1.duplicate()
    assert dc1.header.data_type.name == dc2.header.data_type.name
    assert dc1.header.unit == dc2.header.unit
    assert dc1.header.analysis_period == dc2.header.analysis_period
    assert dc1.header.metadata == dc2.header.metadata
    assert dc1.values == dc2.values


def test_dict_methods():
    """Test the to/from dict methods for discontinuous collections."""
    header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    values = list(xrange(24))
    dc = HourlyDiscontinuousCollection(header, values,
                                       header.analysis_period.datetimes)
    dc_dict = dc.to_dict()
    reconstruced_dc = HourlyDiscontinuousCollection.from_dict(dc_dict)
    assert dc_dict == reconstruced_dc.to_dict()


def test_dict_methods_continuous():
    """Test the to/from dict methods for continuous collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    values = list(xrange(24))
    dc = HourlyContinuousCollection(header, values)
    dc_dict = dc.to_dict()
    reconstruced_dc = HourlyContinuousCollection.from_dict(dc_dict)

    assert dc_dict == reconstruced_dc.to_dict()


def test_filter_collections_by_statement():
    """Test the method to filter collections by conditional statement."""
    header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    values1 = list(xrange(24))
    values2 = [12] * 24
    values3 = list(xrange(12, 36))
    dc1 = HourlyContinuousCollection(header, values1)
    dc2 = HourlyContinuousCollection(header, values2)
    dc3 = HourlyContinuousCollection(header, values3)

    filt_coll = HourlyContinuousCollection.filter_collections_by_statement(
        [dc1, dc2, dc3], 'a >= 12 and c <= 30')

    assert len(filt_coll) == 3
    assert len(filt_coll[0]) == len(filt_coll[1]) == len(filt_coll[2]) == 7
    assert filt_coll[0].values == (12, 13, 14, 15, 16, 17, 18)
    assert filt_coll[1].values == (12, 12, 12, 12, 12, 12, 12)
    assert filt_coll[2].values == (24, 25, 26, 27, 28, 29, 30)
    assert isinstance(filt_coll[0], HourlyDiscontinuousCollection)


def test_is_in_range_data_type():
    """Test the function to check whether values are in range for the data_type."""
    header1 = Header(Temperature(), 'C', AnalysisPeriod())
    header2 = Header(Temperature(), 'K', AnalysisPeriod())
    val1 = [20] * 8760
    val2 = [-300] * 8760
    val3 = [270] * 8760
    val4 = [-10] * 8760
    dc1 = HourlyContinuousCollection(header1, val1)
    dc2 = HourlyContinuousCollection(header1, val2)
    dc3 = HourlyContinuousCollection(header2, val3)
    dc4 = HourlyContinuousCollection(header2, val4)
    assert dc1.is_in_data_type_range(raise_exception=False)
    assert not dc2.is_in_data_type_range(raise_exception=False)
    assert dc3.is_in_data_type_range(raise_exception=False)
    assert not dc4.is_in_data_type_range(raise_exception=False)



# coding=utf-8
from __future__ import division

from ladybug.datatype.temperature import Temperature
from ladybug.analysisperiod import AnalysisPeriod
from ladybug.header import Header
from ladybug.legend import Legend, LegendParameters
from ladybug.datacollection import HourlyContinuousCollection
from ladybug.datacollectionimmutable import HourlyContinuousCollectionImmutable
from ladybug.hourlyplot import HourlyPlot
from ladybug.epw import EPW

from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug_geometry.geometry2d.line import LineSegment2D
from ladybug_geometry.geometry2d.polyline import Polyline2D
from ladybug_geometry.geometry2d.mesh import Mesh2D
from ladybug_geometry.geometry3d.pointvector import Point3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug_geometry.geometry3d.line import LineSegment3D
from ladybug_geometry.geometry3d.polyline import Polyline3D
from ladybug_geometry.geometry3d.mesh import Mesh3D


def test_hourlyplot_init():
    """Test the initialization of HourlyPlot and basic properties."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(range(8760))
    data_coll = HourlyContinuousCollection(header, values)
    hour_plot = HourlyPlot(data_coll)

    str(hour_plot)  # test the string representation
    assert isinstance(hour_plot.data_collection, HourlyContinuousCollectionImmutable)
    assert isinstance(hour_plot.legend_parameters, LegendParameters)
    assert hour_plot.base_point == Point3D(0, 0, 0)
    assert hour_plot.x_dim == 1
    assert hour_plot.y_dim == 4
    assert hour_plot.z_dim == 0
    assert hour_plot.values == data_coll.values

    mesh = hour_plot.colored_mesh2d
    assert isinstance(mesh, Mesh2D)
    assert len(mesh.faces) == 8760
    mesh = hour_plot.colored_mesh3d
    assert isinstance(mesh, Mesh3D)
    assert len(mesh.faces) == 8760

    border = hour_plot.chart_border2d
    assert isinstance(border, Polyline2D)
    assert len(border.segments) == 4
    border = hour_plot.chart_border3d
    assert isinstance(border, Polyline3D)
    assert len(border.segments) == 4

    hour_txt = hour_plot.hour_labels
    assert all(isinstance(txt, str) for txt in hour_txt)
    hour_lines = hour_plot.hour_lines2d
    hour_pts = hour_plot.hour_label_points2d
    assert len(hour_lines) == len(hour_txt) == len(hour_pts) == 5
    assert all(isinstance(line, LineSegment2D) for line in hour_lines)
    assert all(isinstance(pt, Point2D) for pt in hour_pts)
    hour_lines = hour_plot.hour_lines3d
    hour_pts = hour_plot.hour_label_points3d
    assert len(hour_lines) == len(hour_txt) == len(hour_pts) == 5
    assert all(isinstance(line, LineSegment3D) for line in hour_lines)
    assert all(isinstance(pt, Point3D) for pt in hour_pts)

    month_txt = hour_plot.month_labels
    assert all(isinstance(txt, str) for txt in month_txt)
    month_lines = hour_plot.month_lines2d
    month_pts = hour_plot.month_label_points2d
    assert len(month_txt) == len(month_pts) == 12
    assert len(month_lines) == 11
    assert all(isinstance(line, LineSegment2D) for line in month_lines)
    assert all(isinstance(pt, Point2D) for pt in month_pts)
    month_lines = hour_plot.month_lines3d
    month_pts = hour_plot.month_label_points3d
    assert len(month_txt) == len(month_pts) == 12
    assert len(month_lines) == 11
    assert all(isinstance(line, LineSegment3D) for line in month_lines)
    assert all(isinstance(pt, Point3D) for pt in month_pts)

    assert isinstance(hour_plot.legend, Legend)

    assert isinstance(hour_plot.title_text, str)
    assert 'Temperature' in hour_plot.title_text
    assert isinstance(hour_plot.lower_title_location, Plane)
    assert isinstance(hour_plot.upper_title_location, Plane)


def test_hourlyplot_custom():
    """Test the initialization of HourlyPlot with custom properties."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(range(8760))
    data_coll = HourlyContinuousCollection(header, values)
    hour_plot = HourlyPlot(data_coll, LegendParameters(), Point3D(10, 10, 10), 2, 2, 100)

    assert hour_plot.base_point == Point3D(10, 10, 10)
    assert hour_plot.x_dim == 2
    assert hour_plot.y_dim == 2
    assert hour_plot.z_dim == 100

    mesh = hour_plot.colored_mesh3d
    assert isinstance(mesh, Mesh3D)
    assert len(mesh.faces) == 8760
    assert mesh.min == Point3D(10, 10, 10)
    assert mesh.max == Point3D(2 * 365 + 10, 2 * 24 + 10, 110)


def test_hourlyplot_init_from_z_dim_per_unit():
    """Test the initialization of HourlyPlot from_z_dim_per_unit."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(range(8760))
    data_coll = HourlyContinuousCollection(header, values)
    hour_plot = HourlyPlot.from_z_dim_per_unit(
        data_coll, LegendParameters(), Point3D(10, 10, 10), 2, 2, 1)

    assert hour_plot.base_point == Point3D(10, 10, 10)
    assert hour_plot.x_dim == 2
    assert hour_plot.y_dim == 2
    assert hour_plot.z_dim == 8759

    mesh = hour_plot.colored_mesh3d
    assert isinstance(mesh, Mesh3D)
    assert len(mesh.faces) == 8760
    assert mesh.min == Point3D(10, 10, 10)
    assert mesh.max == Point3D(2 * 365 + 10, 2 * 24 + 10, 8759 + 10)


def test_hourlyplot_sub_hourly():
    """Test the initialization of HourlyPlot and basic properties."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(range(8760))
    dc1 = HourlyContinuousCollection(header, values)
    data_coll = dc1.interpolate_to_timestep(4)
    hour_plot = HourlyPlot(data_coll, y_dim=1)

    mesh = hour_plot.colored_mesh2d
    assert isinstance(mesh, Mesh2D)
    assert len(mesh.faces) == 8760 * 4
    assert mesh.min == Point2D(0, 0)
    assert mesh.max == Point2D(365, 4 * 24)


def test_hourlyplot_analysis_period():
    """Test the initialization of HourlyPlot with an analysis period."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(range(8760))
    data_coll = HourlyContinuousCollection(header, values)
    period = AnalysisPeriod(1, 1, 8, 1, 31, 17)
    data_coll = data_coll.filter_by_analysis_period(period)
    hour_plot = HourlyPlot(data_coll, y_dim=1)

    assert hour_plot.analysis_period == period
    mesh = hour_plot.colored_mesh2d
    assert isinstance(mesh, Mesh2D)
    assert len(mesh.faces) == 31 * (17 - 8 + 1)
    assert mesh.min == Point2D(0, 0)
    assert mesh.max == Point2D(31, 17 - 8 + 1)


def test_hourlyplot_reversed_analysis_period():
    """Test the initialization of HourlyPlot with a reversed analysis period."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(range(8760))
    data_coll = HourlyContinuousCollection(header, values)
    period = AnalysisPeriod(12, 1, 0, 1, 31, 23)
    data_coll = data_coll.filter_by_analysis_period(period)
    hour_plot = HourlyPlot(data_coll, y_dim=1)

    assert hour_plot.analysis_period == period
    mesh = hour_plot.colored_mesh2d
    assert isinstance(mesh, Mesh2D)
    assert len(mesh.faces) == 31 * 2 * 24
    assert mesh.min == Point2D(0, 0)
    assert mesh.max == Point2D(31 * 2, 24)


def test_hourlyplot_conditional_statement():
    """Test the initialization of HourlyPlot with a conditional statement."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(range(8760))
    data_coll = HourlyContinuousCollection(header, values)
    data_coll = data_coll.filter_by_conditional_statement('a >= 60')
    hour_plot = HourlyPlot(data_coll, y_dim=1)

    mesh = hour_plot.colored_mesh2d
    assert len(mesh.faces) == 8700
    assert mesh.min == Point2D(0, 0)
    assert mesh.max == Point2D(365, 24)


def test_hourlyplot_custom_hour_labels():
    """Test the initialization of HourlyPlot and basic properties."""
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(range(8760))
    data_coll = HourlyContinuousCollection(header, values)
    hour_plot = HourlyPlot(data_coll)

    custom_labels = [0, 3, 6, 9, 12, 15, 18, 21, 24]

    hour_txt = hour_plot.custom_hour_labels(custom_labels)
    assert all(isinstance(txt, str) for txt in hour_txt)
    hour_lines = hour_plot.custom_hour_lines2d(custom_labels)
    hour_pts = hour_plot.custom_hour_label_points2d(custom_labels)
    assert len(hour_lines) == len(hour_txt) == len(hour_pts) == 9
    assert all(isinstance(line, LineSegment2D) for line in hour_lines)
    assert all(isinstance(pt, Point2D) for pt in hour_pts)
    hour_lines = hour_plot.custom_hour_lines3d(custom_labels)
    hour_pts = hour_plot.custom_hour_label_points3d(custom_labels)
    assert len(hour_lines) == len(hour_txt) == len(hour_pts) == 9
    assert all(isinstance(line, LineSegment3D) for line in hour_lines)
    assert all(isinstance(pt, Point3D) for pt in hour_pts)


def test_hourlyplot_epw():
    """Test the initialization of HourlyPlot with EPW data collections."""
    path = './tests/fixtures/epw/tokyo.epw'
    epw = EPW(path)

    hour_plot = HourlyPlot(epw.dry_bulb_temperature)
    mesh = hour_plot.colored_mesh2d
    assert len(mesh.faces) == 8760
    assert 'Dry Bulb Temperature' in hour_plot.title_text
    assert 'C' in hour_plot.legend.title

    hour_plot = HourlyPlot(epw.global_horizontal_radiation)
    mesh = hour_plot.colored_mesh2d
    assert len(mesh.faces) == 8760
    assert 'Global Horizontal Radiation' in hour_plot.title_text
    assert 'Wh/m2' in hour_plot.legend.title



# coding=utf-8
"""Utility functions for converting between humidity metrics."""
from __future__ import division

import math


def saturated_vapor_pressure(t_kelvin):
    """Saturated vapor pressure (Pa) at a given dry bulb temperature (K).

    This function accounts for the different behavior above vs. below
    the freezing point of water.

    Args:
        t_kelvin: Dry bulb temperature (K).

    Returns:
        Saturated vapor pressure (Pa).

    Note:
        [1] ASHRAE Handbook - Fundamentals (2017) ch. 1 eqn. 5 and 6

        [2] Meyer et al., (2019). PsychroLib: a library of psychrometric
        functions to calculate thermodynamic properties of air. Journal of
        Open Source Software, 4(33), 1137, https://doi.org/10.21105/joss.01137
        https://github.com/psychrometrics/psychrolib/blob/master/src/python/psychrolib.py
    """
    if (t_kelvin <= 273.15):  # saturation vapor pressure below freezing
        ln_p_ws = -5.6745359E+03 / t_kelvin + 6.3925247 - 9.677843E-03 * t_kelvin + \
            6.2215701E-07 * t_kelvin**2 + 2.0747825E-09 * math.pow(t_kelvin, 3) - \
            9.484024E-13 * math.pow(t_kelvin, 4) + 4.1635019 * math.log(t_kelvin)
    else:  # saturation vapor pressure above freezing
        ln_p_ws = -5.8002206E+03 / t_kelvin + 1.3914993 - 4.8640239E-02 * t_kelvin + \
            4.1764768E-05 * t_kelvin**2 - 1.4452093E-08 * math.pow(t_kelvin, 3) + \
            6.5459673 * math.log(t_kelvin)
    return math.exp(ln_p_ws)


def humid_ratio_from_db_rh(db_temp, rel_humid, b_press=101325):
    """Humidity ratio (kg water/kg air) from air temperature (C) and relative humidity (%).

    Args:
        db_temp: Dry bulb temperature (C).
        rel_humid: Relative humidity (%).
        b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

    Returns:
        Humidity ratio (kg water/kg air).

    Note:
        [1] ASHRAE Handbook - Fundamentals (2017) ch. 1 eqn 20

        [2] Meyer et al., (2019). PsychroLib: a library of psychrometric
        functions to calculate thermodynamic properties of air. Journal of
        Open Source Software, 4(33), 1137, https://doi.org/10.21105/joss.01137
        https://github.com/psychrometrics/psychrolib/blob/master/src/python/psychrolib.py
    """
    p_ws = saturated_vapor_pressure(db_temp + 273.15)  # saturation pressure
    p_w = p_ws * (rel_humid / 100)  # partial pressure
    return (p_w * 0.621945) / (b_press - p_w)  # humidity ratio


def enthalpy_from_db_hr(db_temp, humid_ratio, reference_temp=0):
    """Enthalpy (kJ/kg) at a given humidity ratio (water/air) and dry bulb temperature (C).

    Args:
        db_temp: Dry bulb temperature (C).
        humid_ratio: Humidity ratio (kg water/kg air).
        reference_temp: Reference dry air temperature (C). Default is 0C,
            which is standard practice for SI enthalpy values. However, for
            IP enthalpy, this is typically at 0F (-17.78C). Alternatively, for
            absolute thermodynamic enthalpy, one can input 0K (-273.15).

    Returns:
        Enthalpy (kJ/kg).

    Note:
        [1] ASHRAE Handbook - Fundamentals (2017) ch. 1 eqn 30

        [2] Meyer et al., (2019). PsychroLib: a library of psychrometric
        functions to calculate thermodynamic properties of air. Journal of
        Open Source Software, 4(33), 1137, https://doi.org/10.21105/joss.01137
        https://github.com/psychrometrics/psychrolib/blob/master/src/python/psychrolib.py
    """
    correct_temp = db_temp - reference_temp
    enthalpy = 1.006 * correct_temp + humid_ratio * (2501. + 1.86 * correct_temp)
    return enthalpy if enthalpy >= 0 else 0


def dew_point_from_db_rh(db_temp, rel_humid):
    """Dew point temperature (C) from air temperature (C) and relative humidity (%).

    The dew point temperature is solved by inverting the equation giving water vapor
    pressure at saturation from temperature, which is relatively slow but ensures
    high accuracy down to 0.1 C at a wide range of dry bulb temperatures.
    The Newton-Raphson (NR) method is used on the logarithm of water vapour
    pressure as a function of temperature, which is a very smooth function
    Convergence is usually achieved in 3 to 5 iterations.

    Args:
        db_temp: Dry bulb temperature (C).
        rel_humid: Relative humidity (%).

    Returns:
        Dew point temperature (C).

    Note:
        [1] ASHRAE Handbook - Fundamentals (2017) ch. 1 eqn. 5 and 6

        [2] Meyer et al., (2019). PsychroLib: a library of psychrometric
        functions to calculate thermodynamic properties of air. Journal of
        Open Source Software, 4(33), 1137, https://doi.org/10.21105/joss.01137
        https://github.com/psychrometrics/psychrolib/blob/master/src/python/psychrolib.py
    """
    p_ws = saturated_vapor_pressure(db_temp + 273.15)  # saturation pressure
    p_w = p_ws * (rel_humid / 100)  # partial pressure

    # We use NR to approximate the solution.
    td = db_temp  # First guess for dew point temperature (solved for iteratively)
    try:
        ln_vp = math.log(p_w)  # partial pressure of water vapor in moist air
    except ValueError:  # relative humidity of 0, return absolute zero
        return -273.15

    index = 1
    while True:
        td_iter = td   # td used in NR calculation
        ln_vp_iter = math.log(saturated_vapor_pressure(td_iter + 273.15))
        d_ln_vp = _d_ln_p_ws(td_iter)  # Derivative of function, calculated analytically
        td = td_iter - (ln_vp_iter - ln_vp) / d_ln_vp  # New estimate

        if ((math.fabs(td - td_iter) <= 0.1)):  # 0.1 is degree C tolerance
            break  # solution has been found
        if (index > 100):  # 100 is the max iterations (usually only 3-5 are needed)
            break  # max number of iterations has been exceeded
        index = index + 1

    return min(td, db_temp)


def wet_bulb_from_db_rh(db_temp, rel_humid, b_press=101325):
    """Wet bulb temperature (C) from air temperature (C) and relative humidity (%).

    Args:
        db_temp: Dry bulb temperature (C).
        rel_humid: Relative humidity (%).
        b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

    Returns:
        Wet bulb temperature (C).

    Note:
        [1] ASHRAE Handbook - Fundamentals (2017) ch. 1 eqn 33 and 35

        [2] Meyer et al., (2019). PsychroLib: a library of psychrometric
        functions to calculate thermodynamic properties of air. Journal of
        Open Source Software, 4(33), 1137, https://doi.org/10.21105/joss.01137
        https://github.com/psychrometrics/psychrolib/blob/master/src/python/psychrolib.py
    """
    humid_ratio = humid_ratio_from_db_rh(db_temp, rel_humid, b_press)
    # Initial guesses
    wb_temp_sup = db_temp
    wb_temp_inf = dew_point_from_db_rh(db_temp, rel_humid)
    wb_temp = (wb_temp_inf + wb_temp_sup) / 2

    index = 1
    while ((wb_temp_sup - wb_temp_inf) > 0.1):  # 0.1 is degree C tolerance
        # Compute humidity ratio at temperature Tstar
        w_star = humid_ratio_from_db_wb(db_temp, wb_temp, b_press)
        # Get new bounds
        if w_star > humid_ratio:
            wb_temp_sup = wb_temp
        else:
            wb_temp_inf = wb_temp
        # New guess of wet bulb temperature
        wb_temp = (wb_temp_sup + wb_temp_inf) / 2
        if index >= 100:
            break  # 100 is the max iterations (usually only 3-5 are needed)
        index = index + 1
    return wb_temp


def rel_humid_from_db_hr(db_temp, humid_ratio, b_press=101325):
    """Relative Humidity (%) from humidity ratio (water/air) and air temperature (C).

    Args:
        db_temp: Dry bulb temperature (C).
        humid_ratio: Humidity ratio (kg water/kg air).
        b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

    Returns:
        Relative humidity (%).
    """
    pw = (humid_ratio * 1000 * b_press) / (621.9907 + (humid_ratio * 1000))
    pws = saturated_vapor_pressure(db_temp + 273.15)
    return (pw / pws) * 100


def rel_humid_from_db_enth(db_temp, enthalpy, b_press=101325, reference_temp=0):
    """Relative Humidity (%) from air temperature (C) and enthalpy (kJ/kg).

    Args:
        db_temp: Dry bulb temperature (C).
        enthalpy: Enthalpy (kJ/kg).
        b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).
        reference_temp: Reference dry air temperature (C). Default is 0C,
            which is standard practice for SI enthalpy values. However, for
            IP enthalpy, this is typically at 0F (-17.78C). Alternatively, for
            absolute thermodynamic enthalpy, one can input 0K (-273.15).

    Returns:
        Relative humidity (%).
    """
    correct_temp = db_temp - reference_temp
    hr = (enthalpy - (1.006 * correct_temp)) / ((1.86 * correct_temp) + 2501)
    return rel_humid_from_db_hr(db_temp, hr, b_press)


def rel_humid_from_db_dpt(db_temp, dew_pt):
    """Relative humidity (%) from dry bulb temperature (C), and dew point temperature (C).

    Args:
        db_temp: Dry bulb temperature (C).
        dew_pt: Dew point temperature (C).

    Returns:
        Relative humidity (%).
    """
    pws_ta = saturated_vapor_pressure(db_temp + 273.15)
    pws_td = saturated_vapor_pressure(dew_pt + 273.15)
    return 100 * (pws_td / pws_ta)


def rel_humid_from_db_wb(db_temp, wet_bulb, b_press=101325):
    """Relative humidity (%) from dry bulb temperature (C), and wet bulb temperature (C).

    Args:
        db_temp: Dry bulb temperature (C).
        wet_bulb: Wet bulb temperature (C).
        b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

    Returns:
        Relative humidity (%).
    """
    # Calculate saturation pressures
    p_ws = saturated_vapor_pressure(db_temp + 273.15)
    p_ws_wb = saturated_vapor_pressure(wet_bulb + 273.15)
    # calculate partial vapor pressure
    p_w = p_ws_wb - (b_press * 0.000662 * (db_temp - wet_bulb))
    return (p_w / p_ws) * 100


def dew_point_from_db_hr(db_temp, humid_ratio, b_press=101325):
    """Dew Point Temperature (C) from air temperature (C) and humidity ratio (water/air).

    Args:
        db_temp: Dry bulb temperature (C).
        humid_ratio: Humidity ratio (kg water/kg air).
        b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

    Returns:
        Dew point temperature (C).
    """
    rh = rel_humid_from_db_hr(db_temp, humid_ratio, b_press)
    return dew_point_from_db_rh(db_temp, rh)


def dew_point_from_db_enth(db_temp, enthalpy, b_press=101325, reference_temp=0):
    """Dew point temperature (C) from air temperature (C) and enthalpy (kJ/kg).

    Args:
        db_temp: Dry bulb temperature (C).
        enthalpy: Enthalpy (kJ/kg).
        b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).
        reference_temp: Reference dry air temperature (C). Default is 0C,
            which is standard practice for SI enthalpy values. However, for
            IP enthalpy, this is typically at 0F (-17.78C). Alternatively, for
            absolute thermodynamic enthalpy, one can input 0K (-273.15).

    Returns:
        Dew point temperature (C).
    """
    rh = rel_humid_from_db_enth(db_temp, enthalpy, b_press, reference_temp)
    return dew_point_from_db_rh(db_temp, rh)


def dew_point_from_db_wb(db_temp, wet_bulb, b_press=101325):
    """Dew point temperature (C) from dry bulb (C) and wet bulb temperature (C).

    Args:
        db_temp: Dry bulb temperature (C).
        wet_bulb: Wet bulb temperature (C).
        b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

    Returns:
        Dew point temperature (C).
    """
    rh = rel_humid_from_db_wb(db_temp, wet_bulb, b_press)
    return dew_point_from_db_rh(db_temp, rh)


def humid_ratio_from_db_wb(db_temp, wb_temp, b_press=101325):
    """Humidity ratio from air temperature (C) and wet bulb temperature (C).

    Args:
        db_temp: Dry bulb temperature (C).
        wb_temp: Wet bulb temperature (C).
        b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

    Returns:
        Humidity ratio (kg water / kg air).

    Note:
        [1] ASHRAE Handbook - Fundamentals (2017) ch. 1 eqn 36, solved for W
    """
    p_ws = saturated_vapor_pressure(wb_temp + 273.15)
    p_ws_star = 0.621945 * p_ws / (b_press - p_ws)
    if wb_temp >= 0:
        humid_ratio = ((2501. - 2.326 * wb_temp) * p_ws_star - 1.006 * (db_temp - wb_temp)) \
            / (2501. + 1.86 * db_temp - 4.186 * wb_temp)
    else:
        humid_ratio = ((2830. - 0.24 * wb_temp) * p_ws_star - 1.006 * (db_temp - wb_temp)) \
            / (2830. + 1.86 * db_temp - 2.1 * wb_temp)
    return humid_ratio


def db_temp_from_enth_hr(enthalpy, humid_ratio, reference_temp=0):
    """Dry bulb temperature (C) from enthalpy (kJ/kg) and humidity ratio (water/air).

    Args:
        enthalpy: Enthalpy (kJ/kg).
        humid_ratio: Humidity ratio (kg water/kg air).
        reference_temp: Reference dry air temperature (C). Default is 0C,
            which is standard practice for SI enthalpy values. However, for
            IP enthalpy, this is typically at 0F (-17.78C). Alternatively, for
            absolute thermodynamic enthalpy, one can input 0K (-273.15).

    Returns:
        Dry bulb temperature (C).

    Note:
        [1] ASHRAE Handbook - Fundamentals (2017) ch. 1 eqn 30
    """
    db_temp = (enthalpy - 2501. * humid_ratio) / (1.006 + 1.86 * humid_ratio)
    return db_temp + reference_temp


def dew_point_from_db_rh_fast(db_temp, rel_humid):
    """Dew point temperature (C) from air temperature (C) and relative humidity (%).

    Note that the formula here is fast but is only accurate up to 90C. For accurate
    values at extreme temperatures, the dew_point_from_db_rh
    function should be used.

    Args:
        db_temp: Dry bulb temperature (C).
        rel_humid: Relative humidity (%).

    Returns:
        Dew point temperature (C).

    Note:
        [1] J. Sullivan and L. D. Sanders. "Method for obtaining wet-bulb temperatures
        by modifying the psychrometric formula." Center for Experiment Design and Data
        Analysis. NOAA - National Oceanic and Atmospheric Administration.
        https://www.weather.gov/epz/wxcalc_rh
    """
    es = 6.112 * math.e**((17.67 * db_temp) / (db_temp + 243.5))
    e = (es * rel_humid) / 100
    try:
        return (243.5 * math.log(e / 6.112)) / (17.67 - math.log(e / 6.112))
    except ValueError:  # relative humidity of 0, return absolute zero
        return -273.15


def wet_bulb_from_db_rh_fast(db_temp, rel_humid, b_press=101325):
    """Wet bulb temperature (C) from air temperature (C) and relative humidity (%).

    Note that the formula here is fast but is only accurate around temperatures
    of 20C and lower. For accurate values at extreme temperatures, the
    wet_bulb_from_db_rh function should be used.

    Args:
        db_temp: Dry bulb temperature (C).
        rel_humid: Relative humidity (%).
        b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

    Returns:
        Wet bulb temperature (C).

    Note:
        [1] J. Sullivan and L. D. Sanders. "Method for obtaining wet-bulb temperatures
        by modifying the psychrometric formula." Center for Experiment Design and Data
        Analysis. NOAA - National Oceanic and Atmospheric Administration.
        https://www.weather.gov/epz/wxcalc_rh
    """
    es = 6.112 * math.e**((17.67 * db_temp) / (db_temp + 243.5))
    e = (es * rel_humid) / 100
    t_w = 0
    increse = 10.0
    previoussign = 1
    e_d = 1
    while math.fabs(e_d) > 0.005:
        e_wg = 6.112 * (math.e**((17.67 * t_w) / (t_w + 243.5)))
        eg = e_wg - (b_press/100) * (db_temp - t_w) * 0.00066 * (1 + (0.00155 * t_w))
        e_d = e - eg
        if e_d == 0:
            break
        else:
            if e_d < 0:
                cursign = -1
                if cursign != previoussign:
                    previoussign = cursign
                    increse = increse / 10
                else:
                    increse = increse
            else:
                cursign = 1
                if cursign != previoussign:
                    previoussign = cursign
                    increse = increse / 10
                else:
                    increse = increse
        t_w = t_w + increse * previoussign
    return t_w


def _d_ln_p_ws(db_temp):
    """Helper function returning the derivative of the natural log of the
    saturation vapor pressure as a function of dry-bulb temperature.

    Args:
        db_temp : Dry bulb temperature (C).
    Returns:
        Derivative of natural log of vapor pressure of saturated air in Pa.
    """
    T = db_temp + 273.15  # temperature in kelvin
    if db_temp <= 0.:
        d_ln_p_ws = 5.6745359E+03 / math.pow(T, 2) - 9.677843E-03 + 2 * \
            6.2215701E-07 * T + 3 * 2.0747825E-09 * math.pow(T, 2) - 4 * \
            9.484024E-13 * math.pow(T, 3) + 4.1635019 / T
    else:
        d_ln_p_ws = 5.8002206E+03 / math.pow(T, 2) - 4.8640239E-02 + 2 * \
            4.1764768E-05 * T - 3 * 1.4452093E-08 * math.pow(T, 2) + \
            6.5459673 / T
    return d_ln_p_ws



# coding=utf-8
"""Ladybug color, colorsets and colorrange."""
from __future__ import division

try:
    from collections.abc import Iterable  # python < 3.7
except ImportError:
    from collections import Iterable  # python >= 3.8


class Color(object):
    """Ladybug RGB color.

    Args:
        r: red value 0-255, default: 0
        g: green value 0-255, default: 0
        b: blue red value 0-255, default: 0

    Properties:
        * r
        * g
        * b
    """

    __slots__ = ("_r", "_g", "_b")

    def __init__(self, r, g, b):
        """Generate RGB Color.
        """
        self.r = r
        self.g = g
        self.b = b

    @classmethod
    def from_dict(cls, data):
        """Create a color from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

                {
                "r": 255,
                "g": 0,
                "b": 150
                }
        """
        return cls(data['r'], data['g'], data['b'])

    @property
    def r(self):
        """Return R value."""
        return self._r

    @r.setter
    def r(self, value):
        assert 0 <= int(value) <= 255, "%d is out of range. " % value + \
            "R value should be between 0-255"
        self._r = int(value)

    @property
    def g(self):
        """Return G value."""
        return self._g

    @g.setter
    def g(self, value):
        assert 0 <= int(value) <= 255, "%d is out of range. " % value + \
            "G value should be between 0-255"
        self._g = int(value)

    @property
    def b(self):
        """Return B value."""
        return self._b

    @b.setter
    def b(self, value):
        assert 0 <= int(value) <= 255, "%d is out of range. " % value + \
            "B value should be between 0-255"
        self._b = int(value)

    def duplicate(self):
        """Return a copy of the current color."""
        return self.__copy__()

    def to_dict(self):
        """Get color as a dictionary."""
        return {
            'r': self.r,
            'g': self.g,
            'b': self.b,
            'type': 'Color'
        }

    def __copy__(self):
        return self.__class__(self.r, self.g, self.b)

    def __eq__(self, other):
        if isinstance(other, Color):
            return self.r == other.r and self.g == other.g and self.b == other.b
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.r, self.g, self.b))

    def __len__(self):
        return 3

    def __getitem__(self, key):
        return (self.r, self.g, self.b)[key]

    def __iter__(self):
        return iter((self.r, self.g, self.b))

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Return RGB values."""
        return "(R:%d, G:%d, B:%d)" % (self._r, self._g, self._b)


# TODO: Add support for CMYK
class Colorset(object):
    """Ladybug Color-range repository.

    A list of default Ladybug colorsets for color range:
        * 0 - original Ladybug
        * 1 - nuanced Ladybug
        * 2 - Multi-colored Ladybug
        * 3 - View Analysis 1
        * 4 - View Analysis 2 (Red,Green,Blue)
        * 5 - Sunlight Hours
        * 6 - ecotect
        * 7 - thermal Comfort Percentage
        * 8 - thermal Comfort Colors
        * 9 - thermal Comfort Colors (UTCI)
        * 10 - Hot Hours
        * 11 - Cold Hours
        * 12 - Shade Benefit/Harm
        * 13 - thermal Comfort Colors v2 (UTCI)
        * 14 - Shade Harm
        * 15 - Shade Benefit
        * 16 - Black to White
        * 17 - CFD Colors 1
        * 18 - CFD Colors 2
        * 19 - Energy Balance
        * 20 - THERM
        * 21 - Cloud Cover

    Usage:

    .. code-block:: python

        # initialize colorsets
        cs = Colorset()
        print(cs[0])
        >> [<R:75, G:107, B:169>, <R:115, G:147, B:202>, <R:170, G:200, B:247>,
            <R:193, G:213, B:208>, <R:245, G:239, B:103>, <R:252, G:230, B:74>,
            <R:239, G:156, B:21>, <R:234, G:123, B:0>, <R:234, G:74, B:0>,
            <R:234, G:38, B:0>]
    """
    # base color sets for which there are several variations
    _multicolored = [(4, 25, 145), (7, 48, 224), (7, 88, 255), (1, 232, 255),
                     (97, 246, 156), (166, 249, 86), (254, 244, 1), (255, 121, 0),
                     (239, 39, 0), (138, 17, 0)]
    _thermalcomfort = [(0, 136, 255), (200, 225, 255), (255, 255, 255),
                       (255, 230, 230), (255, 0, 0)]
    _benefitharm = [(0, 191, 48), (255, 238, 184), (255, 0, 0)]
    _shadebenefitharm = [(5, 48, 97), (33, 102, 172), (67, 147, 195), (146, 197, 222),
                         (209, 229, 240), (255, 255, 255), (253, 219, 199),
                         (244, 165, 130), (214, 96, 77), (178, 24, 43), (103, 0, 31)]

    # dictionary of all color sets together
    _colors = {
        0: [(75, 107, 169), (115, 147, 202), (170, 200, 247), (193, 213, 208),
            (245, 239, 103), (252, 230, 74), (239, 156, 21), (234, 123, 0),
            (234, 74, 0), (234, 38, 0)],
        1: [(49, 54, 149), (69, 117, 180), (116, 173, 209), (171, 217, 233),
            (224, 243, 248), (255, 255, 191), (254, 224, 144), (253, 174, 97),
            (244, 109, 67), (215, 48, 39), (165, 0, 38)],
        2: _multicolored,
        3: [(0, 0, 255), (53, 0, 202), (107, 0, 148), (160, 0, 95), (214, 0, 41),
            (255, 12, 0), (255, 66, 0), (255, 119, 0), (255, 173, 0), (255, 226, 0),
            (255, 255, 0)],
        4: [(255, 20, 147), (240, 47, 145), (203, 117, 139), (160, 196, 133),
            (132, 248, 129), (124, 253, 132), (96, 239, 160), (53, 217, 203),
            (15, 198, 240), (0, 191, 255)],
        5: [(55, 55, 55), (235, 235, 235)],
        6: [(156, 217, 255), (255, 243, 77), (255, 115, 0), (255, 0, 0), (0, 0, 0)],
        7: [(0, 0, 0), (110, 0, 153), (255, 0, 0), (255, 255, 102), (255, 255, 255)],
        8: _thermalcomfort,
        9: [(0, 136, 255), (67, 176, 255), (134, 215, 255), (174, 228, 255),
            (215, 242, 255), (255, 255, 255), (255, 243, 243), (255, 0, 0)],
        10: _thermalcomfort[2:],
        11: list(reversed(_thermalcomfort[:3])),
        12: _benefitharm,
        13: _benefitharm[1:],
        14: list(reversed(_benefitharm[:2])),
        15: _shadebenefitharm,
        16: _shadebenefitharm[5:],
        17: list(reversed(_shadebenefitharm[:6])),
        18: list(reversed(_multicolored)),
        19: list(reversed(_multicolored)) + [(128, 102, 64)],
        20: [(0, 0, 0), (137, 0, 139), (218, 0, 218), (196, 0, 255), (0, 92, 255),
             (0, 198, 252), (0, 244, 215), (0, 220, 101), (7, 193, 0), (115, 220, 0),
             (249, 251, 0), (254, 178, 0), (253, 77, 0), (255, 15, 15),
             (255, 135, 135), (255, 255, 255)],
        21: [(0, 251, 255), (255, 255, 255), (217, 217, 217), (83, 114, 115)],
        22: [(0, 0, 0), (255, 255, 255)],
        23: [(0, 0, 255), (0, 255, 100), (255, 0, 0)],
        24: [(0, 16, 120), (38, 70, 160), (5, 180, 222), (16, 180, 109),
             (59, 183, 35), (143, 209, 19), (228, 215, 29), (246, 147, 17),
             (243, 74, 0), (255, 0, 0)],
        25: [(69, 92, 166), (66, 128, 167), (62, 176, 168), (78, 181, 137),
             (120, 188, 59), (139, 184, 46), (197, 157, 54), (220, 144, 57),
             (228, 100, 59), (233, 68, 60)],
        26: [(230, 180, 60), (230, 215, 150), (165, 82, 0),
             (128, 20, 20), (255, 128, 128), (64, 128, 128),
             (128, 128, 128), (255, 128, 128), (128, 64, 0),
             (64, 180, 255), (160, 150, 100), (120, 75, 190), (255, 255, 200),
             (0, 128, 0)]
    }

    def __init__(self):
        """Initialize Color-sets."""
        pass

    @classmethod
    def original(cls):
        """Original Ladybug colors."""
        return tuple(Color(*color) for color in cls._colors[0])

    @classmethod
    def nuanced(cls):
        """Nuanced Ladybug colors."""
        return tuple(Color(*color) for color in cls._colors[1])

    @classmethod
    def multi_colored(cls):
        """Multi-colored legend."""
        return tuple(Color(*color) for color in cls._colors[2])

    @classmethod
    def ecotect(cls):
        """Ecotect colors."""
        return tuple(Color(*color) for color in cls._colors[3])

    @classmethod
    def view_study(cls):
        """View analysis colors."""
        return tuple(Color(*color) for color in cls._colors[4])

    @classmethod
    def shadow_study(cls):
        """Shadow study colors (dark to light grey)."""
        return tuple(Color(*color) for color in cls._colors[5])

    @classmethod
    def glare_study(cls):
        """Useful for depicting spatial glare (light blue to yellow, red, black)."""
        return tuple(Color(*color) for color in cls._colors[6])

    @classmethod
    def annual_comfort(cls):
        """Good for annual metrics like UDI and thermal comfort percent."""
        return tuple(Color(*color) for color in cls._colors[7])

    @classmethod
    def thermal_comfort(cls):
        """Thermal comfort colors (blue to white to red)."""
        return tuple(Color(*color) for color in cls._colors[8])

    @classmethod
    def thermal_comfort_utci(cls):
        """Thermal comfort colors weighted for the UTCI scale (heat more than cold)."""
        return tuple(Color(*color) for color in cls._colors[9])

    @classmethod
    def heat_sensation(cls):
        """Red colors for heat sensation."""
        return tuple(Color(*color) for color in cls._colors[10])

    @classmethod
    def cold_sensation(cls):
        """Blue colors for cold sensation."""
        return tuple(Color(*color) for color in cls._colors[11])

    @classmethod
    def benefit_harm(cls):
        """Benefit / harm study colors (red to light yellow to green)."""
        return tuple(Color(*color) for color in cls._colors[12])

    @classmethod
    def harm(cls):
        """Harm colors (light yellow to red)."""
        return tuple(Color(*color) for color in cls._colors[13])

    @classmethod
    def benefit(cls):
        """Benefit colors (light yellow to green)."""
        return tuple(Color(*color) for color in cls._colors[14])

    @classmethod
    def shade_benefit_harm(cls):
        """Shade benefit / harm colors (dark red to white to dark blue)."""
        return tuple(Color(*color) for color in cls._colors[15])

    @classmethod
    def shade_harm(cls):
        """Shade harm colors (white to dark red)."""
        return tuple(Color(*color) for color in cls._colors[16])

    @classmethod
    def shade_benefit(cls):
        """Shade benefit colors (white to dark blue)."""
        return tuple(Color(*color) for color in cls._colors[17])

    @classmethod
    def energy_balance(cls):
        """Energy Balance colors."""
        return tuple(Color(*color) for color in cls._colors[18])

    @classmethod
    def energy_balance_storage(cls):
        """Energy Balance colors with a brown color for storage term."""
        return tuple(Color(*color) for color in cls._colors[19])

    @classmethod
    def therm(cls):
        """THERM colors."""
        return tuple(Color(*color) for color in cls._colors[20])

    @classmethod
    def cloud_cover(cls):
        """Cloud cover colors."""
        return tuple(Color(*color) for color in cls._colors[21])

    @classmethod
    def black_to_white(cls):
        """Black to white colors."""
        return tuple(Color(*color) for color in cls._colors[22])

    @classmethod
    def blue_green_red(cls):
        """Blue to Green to Red colors."""
        return tuple(Color(*color) for color in cls._colors[23])

    @classmethod
    def multicolored_2(cls):
        """Multi-colored colors with less saturation."""
        return tuple(Color(*color) for color in cls._colors[24])

    @classmethod
    def multicolored_3(cls):
        """Multi-colored colors with the least saturation."""
        return tuple(Color(*color) for color in cls._colors[25])

    @classmethod
    def openstudio_palette(cls):
        """Standard color set for the OpenStudio surface types. Ordered as follows.

        Exterior Wall, Interior Wall, Underground Wall,
        Roof, Ceiling, Underground Roof,
        Exposed Floor, Interior Floor, Ground Floor,
        Window, Door, Shade, Air
        """
        return tuple(Color(*color) for color in cls._colors[26])

    def __len__(self):
        """Return length of currently installed color sets."""
        return len(self._colors)

    def __getitem__(self, key):
        """Return one of the color sets."""
        return tuple(Color(*color) for color in self._colors[key])

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Colorset representation."""
        return "{} currently installed Colorsets".format(len(self))


class ColorRange(object):
    """Ladybug Color Range. Used to generate colors from numerical values.

    Args:
        colors: A list of colors. Colors should be input as objects with
            R, G, B values. Default is Ladybug's original colorset.
        domain: A list of at least two numbers to set the lower and upper
            boundary of the color range. This can also be a list of more than
            two values, which can be used to approximate logarithmic or other types
            of color scales. However, the number of values in the domain must
            always be less than or equal to the number of colors.
            Default: [0, 1].
        continuous_colors: Boolean. If True, the colors generated from the
            color range will be in a continuous gradient. If False,
            they will be categorized in incremental groups according to the
            number_of_segments. Default: True for continuous colors.

    Properties:
        * colors
        * continuous_colors
        * domain

    Usage:

    .. code-block:: python

        1.
            color_range = ColorRange(continuous_colors=False)
            color_range.domain = [100, 2000]
            color_range.colors = [Color(75, 107, 169), Color(245, 239, 103),
                Color(234, 38, 0)]
            print(color_range.color(99))
            print(color_range.color(100))
            print(color_range.color(2000))
            print(color_range.color(2001))
            >> (R:75, G:107, B:169)
            >> (R:245, G:239, B:103)
            >> (R:245, G:239, B:103)
            >> (R:234, G:38, B:0)

        2.
            color_range = ColorRange(continuous_colors=False)
            color_range.domain = [100, 2000]
            color_range.colors = [Color(75, 107, 169), Color(245, 239, 103),
                Color(234, 38, 0)]
            color_range.color(300)
            >> (R:245, G:239, B:103)
    """

    def __init__(self, colors=None, domain=None, continuous_colors=True):
        """Initiate Ladybug color range.
        """
        self._continuous_colors = True if continuous_colors is None \
            else continuous_colors
        assert isinstance(self._continuous_colors, bool), \
            "continuous_colors should be a Boolean.\nGot {}.".format(
                type(continuous_colors))
        self._is_domain_set = False
        self.colors = colors
        self.domain = domain

    @classmethod
    def from_dict(cls, data):
        """Create a color range from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "colors": [{'r': 0, 'g': 0, 'b': 0}, {'r': 255, 'g': 255, 'b': 255}],
            "domain": [0, 100],
            "continuous_colors": True
            }
        """
        optional_keys = ('colors', 'domain', 'continuous_colors')
        for key in optional_keys:
            if key not in data:
                data[key] = None
        colors = None
        if data['colors'] is not None:
            colors = [Color.from_dict(col) for col in data['colors']]

        return cls(colors, data['domain'], data['continuous_colors'])

    @property
    def colors(self):
        """Get or set the colors defining the color range."""
        return self._colors

    @colors.setter
    def colors(self, cols):
        if not cols:
            self._colors = Colorset.original()
        else:
            assert isinstance(cols, Iterable) \
                and not isinstance(cols, (str, dict, bytes, bytearray)), \
                'Colors should be a list or tuple. Got {}'.format(type(cols))
            try:
                cols = tuple(col if isinstance(col, Color) else Color(
                    col.R, col.G, col.B) for col in cols)
            except AttributeError:
                try:
                    cols = tuple(Color(col.Red, col.Green, col.Blue) for col in cols)
                except AttributeError:
                    raise ValueError("{} is not a valid list of colors".format(cols))
            if self._is_domain_set:
                self.domain = self.domain  # re-check the domain against new colors
            self._colors = cols

    @property
    def domain(self):
        """Get or set the domain defining the color range."""
        return self._domain

    @domain.setter
    def domain(self, dom):
        # check and prepare domain
        if not dom:
            dom = (0, 1)
        else:
            assert isinstance(dom, Iterable) \
                and not isinstance(dom, (str, dict, bytes, bytearray)), \
                'Domain should be a list or tuple. Got {}'.format(type(dom))
            for val in dom:
                assert isinstance(val, (float, int)), 'Values of a domain must be ' \
                    'numbers. Got {}.'.format(type(val))
            dom = sorted(map(float, dom))

        if self._continuous_colors:  # continuous
            # if type is continuous domain can only be 2 values
            # or at least 1 value less than number of colors
            if len(dom) == 2:
                # remap domain based on colors
                _step = float(dom[1] - dom[0]) / (len(self._colors) - 1)
                _n = dom[0]
                dom = tuple(_n + c * _step for c in range(len(self._colors)))
            else:
                assert len(self._colors) >= len(dom), \
                    "For a continuous color range, the length of the domain should " \
                    "be 2 or greater than the number of colors."
        else:  # segmented
            # Number of colors should be at least one more than number of domain values
            assert len(self._colors) > len(dom), \
                "For a segmented color range, the length of colors " + \
                "should be more than the number of domain values ."

        self._is_domain_set = True
        self._domain = tuple(dom)

    @property
    def continuous_colors(self):
        """Boolean noting whether colors generated are continuous or discrete."""
        return self._continuous_colors

    def color(self, value):
        """Calculate a color along the range for an input value."""
        if value < self._domain[0]:
            return self._colors[0]
        if value > self._domain[-1]:
            return self._colors[-1]

        # find the index of the value in domain
        for count, d in enumerate(self._domain):
            if d <= value <= self._domain[count + 1]:
                if self._continuous_colors:
                    return self._cal_color(value, count)
                else:
                    return self._colors[count + 1]

    def duplicate(self):
        """Return a copy of the current color range."""
        return self.__copy__()

    def to_dict(self):
        """Get color range as a dictionary."""
        return {
            'colors': [col.to_dict() for col in self.colors],
            'domain': self.domain,
            'continuous_colors': self.continuous_colors,
            'type': 'ColorRange'
        }

    def _cal_color(self, value, color_index):
        """Blend between two colors based on input value."""
        range_min_p = self._domain[color_index]
        range_p = self._domain[color_index + 1] - range_min_p
        try:
            factor = (value - range_min_p) / range_p
        except ZeroDivisionError:
            factor = 0

        min_color = self.colors[color_index]
        max_color = self.colors[color_index + 1]
        red = round(factor * (max_color.r - min_color.r) + min_color.r)
        green = round(factor * (max_color.g - min_color.g) + min_color.g)
        blue = round(factor * (max_color.b - min_color.b) + min_color.b)

        return Color(red, green, blue)

    def __copy__(self):
        return self.__class__(self.colors, self.domain, self.continuous_colors)

    def __len__(self):
        """Return length of colors."""
        return len(self._colors)

    def __getitem__(self, key):
        """Return key item from the color list."""
        return self._colors[key]

    def __iter__(self):
        """Use colors to iterate."""
        return iter(self._colors)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Color Range representation."""
        return "Color Range ({} colors) (domain {})".format(len(self), self.domain)
